# ruff: noqa: ERA001, PERF203
import dataclasses
from collections.abc import Iterator
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Optional, Union, get_args, get_origin

from typing_extensions import override

from ingeniamotion.fsoe_master.fsoe import (
    FSoEDictionaryItem,
    FSoEDictionaryItemInput,
    FSoEDictionaryItemInputOutput,
)
from ingeniamotion.fsoe_master.parameters import SafetyParameter

if TYPE_CHECKING:
    from ingeniamotion.fsoe_master.handler import FSoEMasterHandler

__all__ = [
    "SafeHomingFunction",
    "SafeInputsFunction",
    "SafetyFunction",
    "SLPFunction",
    "SLSFunction",
    "SOSFunction",
    "SOutFunction",
    "SPFunction",
    "SS1Function",
    "SS2Function",
    "SSRFunction",
    "STOFunction",
    "SVFunction",
    "SDIFunction",
    "SLIFunction",
]


@dataclass(frozen=True)
class SafetyFieldMetadata:
    """Metadata for safety fields.

    Can either represent a prototype for a field (with placeholders in the UID),
    or a specific instance (with concrete values in the UID).

    Attributes:
        uid: Unique identifier for the safety field.
            May include an instance index, e.g. `FSOE_SS1_{i}`.
        display_name: Name of the safety field for documentation and display.
        attr_name: Attribute name of the field in the SafetyFunction class.
    """

    uid: str
    display_name: str
    attr_name: str


def safety_field(uid: str, display_name: str):  # type: ignore[no-untyped-def]
    """Create a dataclass field with metadata for safety functions.

    Args:
        uid: Unique identifier of the register associated with the field.
            May include an instance index, e.g. `FSOE_SS1_{i}`.

        display_name: Name of the field for documentation and display.

    Returns:
        dataclass field with safety metadata
    """
    return dataclasses.field(metadata={"uid": uid, "display_name": display_name})


def is_optional(field_type: type) -> tuple[bool, type]:
    """Check if a field type is Optional and get the internal type.

    Returns:
        A tuple with a boolean indicating if the field is Optional,
        and the internal type if it is Optional, or the original type if not.
    """
    is_optional = get_origin(field_type) is Optional or (
        get_origin(field_type) is Union and type(None) in get_args(field_type)
    )

    if not is_optional:
        return False, field_type

    internal_type = next(x for x in get_args(field_type) if x is not None)
    return True, internal_type


@dataclass()
class SafetyFunction:
    """Base class for Safety Functions.

    Wraps input/output items and parameters used by the FSoE Master handler.
    """

    n_instance: Optional[int]
    """Number of instances of the safety function, if applicable."""

    name: str
    """Name of the safety function.
    In the class var it may include an instance index, e.g. `Safe Stop {i}`."""

    ios: dict[SafetyFieldMetadata, FSoEDictionaryItem]
    parameters: dict[SafetyFieldMetadata, SafetyParameter]
    handler: "FSoEMasterHandler"

    def activated_by(self) -> list["SafetyFunction"]:
        """Get the safety function that activates this function, if any.

        Returns:
            The safety function that activates this function, or None if not activated by any.
        """
        return []

    @classmethod
    def for_handler(cls, handler: "FSoEMasterHandler") -> Iterator["SafetyFunction"]:
        """Get the safety function instances for a given FSoE master handler.

        Yields:
            All safety function instances available for the handler.
        """
        yield from STOFunction._explore_instances(handler)
        yield from SS1Function._explore_instances(handler)
        yield from SafeInputsFunction._explore_instances(handler)
        yield from SOSFunction._explore_instances(handler)
        yield from SS2Function._explore_instances(handler)
        yield from SOutFunction._explore_instances(handler)
        yield from SPFunction._explore_instances(handler)
        yield from SVFunction._explore_instances(handler)
        yield from SafeHomingFunction._explore_instances(handler)
        yield from SLSFunction._explore_instances(handler)
        yield from SSRFunction._explore_instances(handler)
        yield from SLPFunction._explore_instances(handler)
        yield from SDIFunction._explore_instances(handler)
        yield from SLIFunction._explore_instances(handler)

    @classmethod
    def _create_instance(
        cls, handler: "FSoEMasterHandler", n_instance: Optional[int] = None
    ) -> Iterator["SafetyFunction"]:
        """Create an instance of the safety function.

        Args:
            handler: The FSoE master handler to use.
            n_instance: The instance index to use for the safety function.
                Formatted into the UID if provided.
                If not the function is assumed to be single-instance.

        Raises:
            KeyError: If the required input/output items
                or parameters are not found in the handler's dictionary.

        Yields:
            Iterator[SafetyFunction]: An iterator yielding the safety function instance.
        """
        ios: dict[SafetyFieldMetadata, Optional[FSoEDictionaryItem]] = {}
        parameters: dict[SafetyFieldMetadata, Optional[SafetyParameter]] = {}
        for field in dataclasses.fields(cls):
            metadata_dict = field.metadata.copy()
            if "uid" not in metadata_dict:
                continue

            if n_instance is not None:
                metadata_dict["uid"] = metadata_dict["uid"].format(i=n_instance)

            metadata = SafetyFieldMetadata(**metadata_dict, attr_name=field.name)

            optional, field_type = is_optional(field.type)

            if field_type == FSoEDictionaryItemInputOutput:
                ios[metadata] = cls._get_input_output(handler, metadata.uid, optional)
            elif field_type == FSoEDictionaryItemInput:
                ios[metadata] = cls._get_input(handler, metadata.uid, optional)
            elif field_type == SafetyParameter:
                parameters[metadata] = cls._get_parameter(handler, metadata.uid, optional)

        name = cls.name.format(i=n_instance) if n_instance else cls.name

        yield cls(
            n_instance=n_instance,
            name=name,
            ios={metadata: io for metadata, io in ios.items() if io is not None},
            parameters={
                metadata: parameter
                for metadata, parameter in parameters.items()
                if parameter is not None
            },
            handler=handler,
            **{field.attr_name: io for field, io in ios.items()},
            **{field.attr_name: parameter for field, parameter in parameters.items()},
        )

    @classmethod
    def _explore_instances(cls, handler: "FSoEMasterHandler") -> Iterator["SafetyFunction"]:
        """Explore instances of the safety function.

        Tries to create instances of each safety functions according to the availability of
        the input/output items and parameters in the handler's dictionary.

        Yields:
            int: Instances of the safety function available for the handler.
        """
        # Check if the instance is single-instance
        try:
            yield from cls._create_instance(handler)
            return
        except KeyError:
            pass

        # If not single-instance, explore instances until a KeyError is raised.
        i = 1
        try:
            while True:
                yield from cls._create_instance(handler, n_instance=i)
                i += 1
        except KeyError:
            return

    @classmethod
    def _get_input_output(
        cls, handler: "FSoEMasterHandler", uid: str, optional: bool
    ) -> Optional[FSoEDictionaryItemInputOutput]:
        """Get the input/output item from the handler's dictionary.

        Args:
            handler: The FSoE master handler to use.
            uid: The unique identifier of the input/output item.
            optional: Whether the item is optional or required.

        Raises:
            KeyError: if the item is not found.
            TypeError: if the item is not of type FSoEDictionaryItemInputOutput.

        Returns:
            FSoEDictionaryItemInputOutput: The required input/output item.
        """
        item = handler.dictionary.name_map.get(uid)
        if not optional:
            if item is None:
                raise KeyError(f"Dictionary item {uid} not found in the handler's dictionary")
            if not isinstance(item, FSoEDictionaryItemInputOutput):
                raise TypeError(
                    f"Expected DictionaryItemInputOutput {uid} "
                    f"on the safe dictionary, got {type(item)}"
                )
        return item

    @classmethod
    def _get_input(
        cls, handler: "FSoEMasterHandler", uid: str, optional: bool
    ) -> Optional[FSoEDictionaryItemInput]:
        """Get the input item from the handler's dictionary.

        Args:
            handler: The FSoE master handler to use.
            uid: The unique identifier of the input item.
            optional: Whether the item is optional or required.

        Raises:
            KeyError: if the item is not found.
            TypeError: if the item is not of type FSoEDictionaryItemInput.

        Returns:
            FSoEDictionaryItemInput: The required input item.
        """
        item = handler.dictionary.name_map.get(uid)
        if not optional:
            if item is None:
                raise KeyError(f"Dictionary item {uid} not found in the handler's dictionary")
            if not isinstance(item, FSoEDictionaryItemInput):
                raise TypeError(
                    f"Expected DictionaryItemInput {uid} on the safe dictionary, got {type(item)}"
                )
        return item

    @classmethod
    def _get_parameter(
        cls, handler: "FSoEMasterHandler", uid: str, optional: bool
    ) -> Optional[SafetyParameter]:
        """Get the parameter from the handler's safety parameters.

        Args:
            handler: The FSoE master handler to use.
            uid: The unique identifier of the safety parameter.
            optional: Whether the parameter is optional or required.

        Raises:
             KeyError: if the parameter is not found.

        Returns:
                SafetyParameter: The required safety parameter.
        """
        param = handler.safety_parameters.get(uid, None)
        if not optional and uid not in handler.safety_parameters:
            raise KeyError(f"Safety parameter {uid} not found in the handler's safety parameters")

        return param


@dataclass()
class STOFunction(SafetyFunction):
    """Safe Torque Off Safety Function."""

    name = "Safe Torque Off"

    COMMAND_UID = "FSOE_STO"

    command: FSoEDictionaryItemInputOutput = safety_field(uid=COMMAND_UID, display_name="Command")
    activate_sout: Optional[SafetyParameter] = safety_field(
        uid="FSOE_STO_ACTIVATE_SOUT", display_name="Activate SOUT"
    )

    class ActiveSOUT(IntEnum):
        """Enum for the Activate SOUT parameter in the STO function."""

        DISABLED = 0
        ENABLED = 0x66600001


@dataclass()
class SS1Function(SafetyFunction):
    """Safe Stop 1 Safety Function."""

    name = "Safe Stop 1 instance {i}"

    COMMAND_UID = "FSOE_SS1_{i}"

    command: FSoEDictionaryItemInputOutput = safety_field(uid=COMMAND_UID, display_name="Command")
    time_to_sto: SafetyParameter = safety_field(
        uid="FSOE_SS1_TIME_TO_STO_{i}", display_name="Time to STO"
    )
    velocity_zero_window: Optional[SafetyParameter] = safety_field(
        uid="FSOE_SS1_VEL_ZERO_WINDOW_{i}", display_name="Velocity Zero Window"
    )
    time_for_velocity_zero: Optional[SafetyParameter] = safety_field(
        uid="FSOE_SS1_TIME_FOR_VEL_ZERO_{i}", display_name="Time for Velocity Zero"
    )

    time_delay_for_deceleration: Optional[SafetyParameter] = safety_field(
        "FSOE_SS1_TIME_DELAY_DEC_{i}", display_name="Time Delay for Deceleration"
    )
    deceleration_limit: Optional[SafetyParameter] = safety_field(
        uid="FSOE_SS1_DEC_LIMIT_{i}", display_name="Deceleration Limit"
    )
    activate_sout: Optional[SafetyParameter] = safety_field(
        uid="FSOE_SS1_ACTIVATE_SOUT_{i}", display_name="Activate SOUT"
    )

    class ActiveSOUT(IntEnum):
        """Enum for the Activate SOUT parameter in the STO function."""

        DISABLED = 0
        ENABLED = 0x66600001

    @override
    def activated_by(self) -> list[SafetyFunction]:
        sf_list: list[SafetyFunction] = []
        process_image = self.handler.process_image
        si_function: SafeInputsFunction = self.handler.safe_inputs_function()
        if si_function.map.get() == SafeInputsFunction.SafeInputMap.SS1:
            sf_list.append(si_function)
        sf_list.extend(
            slp_function
            for slp_function in self.handler.safety_functions_by_type().get(SLPFunction, [])
            if (
                process_image.is_safety_function_mapped(slp_function, strict=True)
                or slp_function.activated_by() != []
            )
            and slp_function.error_reaction.get() == SLPFunction.ErrorReaction.SS1
        )
        sf_list.extend(
            ssr_function
            for ssr_function in self.handler.safety_functions_by_type().get(SSRFunction, [])
            if (
                process_image.is_safety_function_mapped(ssr_function, strict=True)
                or ssr_function.activated_by() != []
            )
            and ssr_function.error_reaction.get() == SSRFunction.ErrorReaction.SS1
        )
        sf_list.extend(
            sls_function
            for sls_function in self.handler.safety_functions_by_type().get(SLSFunction, [])
            if (
                process_image.is_safety_function_mapped(sls_function, strict=True)
                or sls_function.activated_by() != []
            )
            and sls_function.error_reaction.get() == SLSFunction.ErrorReaction.SS1
        )
        sf_list.extend(
            sli_function
            for sli_function in self.handler.safety_functions_by_type().get(SLIFunction, [])
            if (
                process_image.is_safety_function_mapped(sli_function, strict=True)
                or sli_function.activated_by() != []
            )
            and sli_function.error_reaction.get() == SLIFunction.ErrorReaction.SS1
        )
        return sf_list


@dataclass()
class SafeInputsFunction(SafetyFunction):
    """Safe Inputs Safety Function."""

    name = "Safe Inputs"

    SAFE_INPUTS_UID = "FSOE_SAFE_INPUTS_VALUE"
    value: FSoEDictionaryItemInput = safety_field(uid=SAFE_INPUTS_UID, display_name="Value")
    map: SafetyParameter = safety_field(uid="FSOE_SAFE_INPUTS_MAP", display_name="Map")

    class SafeInputMap(IntEnum):
        """Enum for the Safe Inputs map parameter."""

        NONE = 0
        STO = 1
        SS1 = 2
        SS2 = 3
        SOUT = 4


@dataclass()
class SOSFunction(SafetyFunction):
    """Safe Operation Stop Safety Function."""

    name = "Safe Operation Stop instance {i}"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SOS_{i}", display_name="Command"
    )
    position_zero_window: SafetyParameter = safety_field(
        uid="FSOE_SOS_POS_ZERO_WINDOW_{i}", display_name="Position Zero Window"
    )
    velocity_zero_window: SafetyParameter = safety_field(
        uid="FSOE_SOS_VEL_ZERO_WINDOW_{i}", display_name="Velocity Zero Window"
    )

    @override
    def activated_by(self) -> list[SafetyFunction]:
        sf_list: list[SafetyFunction] = []
        process_image = self.handler.process_image
        ss2_instance = self.handler.get_function_instance(SS2Function)
        if (
            process_image.is_safety_function_mapped(ss2_instance, strict=True)
            or ss2_instance.activated_by() != []
        ):
            sf_list.append(ss2_instance)
        return sf_list


@dataclass()
class SS2Function(SafetyFunction):
    """Safe Stop 2 Safety Function."""

    name = "Safe Stop 2 instance {i}"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SS2_{i}", display_name="Command"
    )
    time_for_velocity_zero: SafetyParameter = safety_field(
        uid="FSOE_SS2_TIME_FOR_VEL_ZERO_{i}", display_name="Time for Velocity Zero"
    )
    time_to_sos: SafetyParameter = safety_field(
        uid="FSOE_SS2_TIME_TO_SOS_{i}", display_name="Time to SOS"
    )
    deceleration_limit: SafetyParameter = safety_field(
        uid="FSOE_SS2_DEC_LIMIT_{i}", display_name="Deceleration Limit"
    )
    time_delay_deceleration_limit: SafetyParameter = safety_field(
        uid="FSOE_SS2_TIME_DELAY_DEC_{i}", display_name="Time Delay Deceleration Limit"
    )
    error_reaction: SafetyParameter = safety_field(
        uid="FSOE_SS2_ERROR_REACTION_{i}", display_name="Error Reaction"
    )

    class ErrorReaction(IntEnum):
        """Enum for the Error Reaction parameter in the SS2 function."""

        NONE = 0
        STO = 0x66400001

    __FSOE_SLP_ERROR_REACTION_ACTIVE_VALUE = 0x66700101

    @override
    def activated_by(self) -> list[SafetyFunction]:
        sf_list: list[SafetyFunction] = []
        process_image = self.handler.process_image
        si_function: SafeInputsFunction = self.handler.safe_inputs_function()
        if si_function.map.get() == SafeInputsFunction.SafeInputMap.SS2:
            sf_list.append(si_function)
        sf_list.extend(
            slp_function
            for slp_function in self.handler.safety_functions_by_type().get(SLPFunction, [])
            if (
                process_image.is_safety_function_mapped(slp_function, strict=True)
                or slp_function.activated_by() != []
            )
            and slp_function.error_reaction.get() == SLPFunction.ErrorReaction.SS2
        )
        return sf_list


@dataclass()
class SOutFunction(SafetyFunction):
    """Safe Output Safety Function."""

    name = "Safe Output"

    command: FSoEDictionaryItemInputOutput = safety_field(uid="FSOE_SOUT", display_name="Command")
    brake_time_delay: SafetyParameter = safety_field(
        uid="FSOE_SOUT_BRAKE_TIME_DELAY", display_name="Electromechanical delay for a brake"
    )
    sout_disable: SafetyParameter = safety_field(
        uid="FSOE_SOUT_DISABLE", display_name="Disables the SOUT functionality"
    )

    @override
    def activated_by(self) -> list[SafetyFunction]:
        sf_list: list[SafetyFunction] = []
        process_image = self.handler.process_image
        sto_function: STOFunction = self.handler.sto_function()
        if (
            sto_function.activate_sout is not None
            and sto_function.activate_sout.get() == STOFunction.ActiveSOUT.ENABLED
        ):
            sf_list.append(sto_function)
        si_function: SafeInputsFunction = self.handler.safe_inputs_function()
        if si_function.map.get() == SafeInputsFunction.SafeInputMap.SOUT:
            sf_list.append(si_function)
        ss1_function: SS1Function = self.handler.ss1_function()
        if (
            (
                process_image.is_safety_function_mapped(ss1_function)
                or ss1_function.activated_by() != []
            )
            and ss1_function.activate_sout is not None
            and ss1_function.activate_sout.get() == SS1Function.ActiveSOUT.ENABLED
        ):
            sf_list.append(ss1_function)
        return sf_list


@dataclass()
class SPFunction(SafetyFunction):
    """Safe Position Safety Function."""

    name = "Safe Position"

    value: FSoEDictionaryItemInput = safety_field(uid="FSOE_SAFE_POSITION", display_name="Value")
    tolerance: SafetyParameter = safety_field(
        uid="FSOE_POSITION_TOLERANCE", display_name="Tolerance"
    )


@dataclass()
class SVFunction(SafetyFunction):
    """Safe Velocity Safety Function."""

    name = "Safe Velocity"

    value: FSoEDictionaryItemInput = safety_field("FSOE_SAFE_VELOCITY", display_name="Value")


@dataclass()
class SafeHomingFunction(SafetyFunction):
    """Safe Homing Safety Function."""

    name = "Safe Homing"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SAFE_HOMING", display_name="Command"
    )
    homing_ref: SafetyParameter = safety_field(
        uid="FSOE_SAFE_HOMING_REFERENCE", display_name="Reference"
    )


@dataclass()
class SLSFunction(SafetyFunction):
    """Safe Limited Speed Safety Function."""

    name = "Safe Limited Speed instance {i}"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SLS_CMD_{i}", display_name="Command"
    )
    velocity_limit: SafetyParameter = safety_field(
        uid="FSOE_SLS_VELOCITY_LIMIT_{i}", display_name="Limit"
    )
    error_reaction: SafetyParameter = safety_field(
        uid="FSOE_SLS_ERROR_REACTION_{i}", display_name="Error Reaction"
    )

    class ErrorReaction(IntEnum):
        """Enum for the Error Reaction parameter in the SLS function."""

        NONE = 0
        STO = 0x66400001
        SS1 = 0x66500101


@dataclass()
class SSRFunction(SafetyFunction):
    """Safe Speed Range Safety Function."""

    name = "Safe Speed Range instance {i}"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SSR_COMMAND_{i}", display_name="Command"
    )
    upper_limit: SafetyParameter = safety_field(
        uid="FSOE_SSR_UPPER_LIMIT_{i}", display_name="Upper Limit"
    )
    lower_limit: SafetyParameter = safety_field(
        uid="FSOE_SSR_LOWER_LIMIT_{i}", display_name="Lower Limit"
    )
    error_reaction: SafetyParameter = safety_field(
        uid="FSOE_SSR_ERROR_REACTION_{i}", display_name="Error Reaction"
    )

    class ErrorReaction(IntEnum):
        """Enum for the Error Reaction parameter in the SSR function."""

        NONE = 0
        STO = 0x66400001
        SS1 = 0x66500101


@dataclass()
class SLPFunction(SafetyFunction):
    """Safe Limited Position Safety Function."""

    name = "Safe Limited Position instance {i}"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SLP_COMMAND_{i}", display_name="Command"
    )
    upper_limit: SafetyParameter = safety_field(
        uid="FSOE_SLP_UPPER_LIMIT_{i}", display_name="Upper Limit"
    )
    lower_limit: SafetyParameter = safety_field(
        uid="FSOE_SLP_LOWER_LIMIT_{i}", display_name="Lower Limit"
    )
    error_reaction: SafetyParameter = safety_field(
        uid="FSOE_SLP_ERROR_REACTION_{i}", display_name="Error Reaction"
    )

    class ErrorReaction(IntEnum):
        """Enum for the Error Reaction parameter in the SLI function."""

        NONE = 0
        STO = 0x66400001
        SS1 = 0x66500101
        SS2 = 0x66700101


@dataclass()
class SDIFunction(SafetyFunction):
    """Safe Direction Safety Function."""

    name = "Safe Direction"

    command_positive: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SDI_P", display_name="Command Positive"
    )
    command_negative: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SDI_N", display_name="Command Negative"
    )
    pos_zero_window: SafetyParameter = safety_field(
        uid="FSOE_SDI_POS_ZERO_WINDOW", display_name="Position Zero Window"
    )


@dataclass()
class SLIFunction(SafetyFunction):
    """Safe Limited Increment Safety Function."""

    name = "Safe Limited Increment instance {i}"

    command: FSoEDictionaryItemInputOutput = safety_field(
        uid="FSOE_SLI_COMMAND_{i}", display_name="Command"
    )
    upper_limit: SafetyParameter = safety_field(
        uid="FSOE_SLI_UPPER_LIMIT_{i}", display_name="Upper Limit"
    )
    lower_limit: SafetyParameter = safety_field(
        uid="FSOE_SLI_LOWER_LIMIT_{i}", display_name="Lower Limit"
    )
    error_reaction: SafetyParameter = safety_field(
        uid="FSOE_SLI_ERROR_REACTION_{i}", display_name="Error Reaction"
    )

    class ErrorReaction(IntEnum):
        """Enum for the Error Reaction parameter in the SLI function."""

        NONE = 0
        STO = 0x66400001
        SS1 = 0x66500101



import inspect
from functools import wraps
from typing import Callable, TypeVar

from ingeniamotion.exceptions import IMStatusWordError

DEFAULT_SERVO = "default"
DEFAULT_AXIS = 1

T = TypeVar("T")


class MCMetaClass(type):
    """MotionController submodules metaclass.

    This class has decorators that can be useful for some
    functions, as motor disabled checker.
    """

    SERVO_ARG_NAME = "servo"
    AXIS_ARG_NAME = "axis"

    @classmethod
    def check_motor_disabled(cls, func: Callable[..., T]) -> Callable[..., T]:
        """Decorator to check if motor is disabled.

        If motor is enabled raises an exception.

        Returns:
            Callable: The wrapped function that checks if the motor is disabled
            before executing the original function.
        """

        @wraps(func)
        def wrapper(self, *args, **kwargs):  # type: ignore[no-untyped-def]
            mc = self.mc
            func_args = inspect.getfullargspec(func).args
            servo_index = func_args.index(cls.SERVO_ARG_NAME)
            axis_index = func_args.index(cls.AXIS_ARG_NAME)
            if len(args) < servo_index:
                servo = kwargs.get(cls.SERVO_ARG_NAME, DEFAULT_SERVO)
            else:
                servo = args[servo_index - 1]
            if len(args) < axis_index:
                axis = kwargs.get(cls.AXIS_ARG_NAME, DEFAULT_AXIS)
            else:
                axis = args[axis_index - 1]
            if mc.configuration.is_motor_enabled(servo=servo, axis=axis):
                raise IMStatusWordError("Motor is enabled")
            return func(self, *args, **kwargs)

        return wrapper



import os
from typing import TYPE_CHECKING, Optional, Union

import ingenialogger
from ingenialink import CanBaudrate
from ingenialink.canopen.network import CanopenNetwork
from ingenialink.dictionary import SubnodeType
from ingenialink.enums.register import RegAccess, RegDtype
from ingenialink.eoe.network import EoENetwork
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.register import Register

from ingeniamotion.enums import CommunicationType
from ingeniamotion.exceptions import IMError, IMRegisterNotExistError
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

logger = ingenialogger.get_logger(__name__)


class Information:
    """Information."""

    PRODUCT_CODE_COMKIT = 0x3214001
    PART_NUMBER_COMKIT = "COM-KIT"

    def __init__(self, motion_controller: "MotionController"):
        self.mc = motion_controller

    def register_info(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> Register:
        """Return register object.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register object.

        Raises:
            IMRegisterNotExistError: If register does not exist in dictionary.

        """
        drive = self.mc._get_drive(servo)
        try:
            return drive.dictionary.registers(axis)[register]
        except KeyError:
            raise IMRegisterNotExistError(
                f"Register: {register} axis: {axis} not exist in dictionary"
            )

    def register_type(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> RegDtype:
        """Return register dtype.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register dtype.

        Raises:
            IMRegisterNotExistError: If register does not exist in dictionary.

        """
        register_obj = self.register_info(register, axis=axis, servo=servo)
        return register_obj.dtype

    def register_access(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> RegAccess:
        """Return register access.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register access.

        Raises:
            IMRegisterNotExistError: If register does not exist in dictionary.

        """
        register_obj = self.register_info(register, axis=axis, servo=servo)
        return register_obj.access

    def register_range(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> Union[tuple[None, None], tuple[int, int], tuple[float, float], tuple[str, str]]:
        """Return register range.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register range, minimum and maximum.

        Raises:
            IMRegisterNotExistError: If register does not exist in dictionary.

        """
        register_obj = self.register_info(register, axis=axis, servo=servo)
        return register_obj.range

    def register_exists(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> bool:
        """Check if register exists in dictionary.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            ``True`` if register exists, else ``False``.

        """
        drive = self.mc._get_drive(servo)
        return register in drive.dictionary.registers(axis)

    def get_product_name(self, servo: str = DEFAULT_SERVO) -> Optional[str]:
        """Get the product name of the drive.

        Args:
            servo: alias of the servo.

        Returns:
            If it exists for example: "EVE-NET-E", "CAP-NET-E", etc.
        """
        drive = self.mc._get_drive(servo)
        product_name = drive.dictionary.part_number
        if self.is_comkit(servo):
            return f"{product_name} + {self.PART_NUMBER_COMKIT}"
        return product_name

    def get_node_id(self, servo: str = DEFAULT_SERVO) -> int:
        """Get the node ID for CANopen communications.

        Args:
            servo: alias of the servo.

        Raises:
            IMError: If the servo is not a CANopen device.

        Returns:
            Node ID of the drive.
        """
        drive = self.mc._get_drive(servo)
        net = self.mc._get_network(servo)
        if isinstance(net, CanopenNetwork):
            return int(drive.target)
        else:
            raise IMError("You need a CANopen communication to use this function")

    def get_baudrate(self, servo: str = DEFAULT_SERVO) -> CanBaudrate:
        """Get the baudrate of target servo.

        Args:
            servo: alias of the servo.

        Raises:
            IMError: If the servo is not a CANopen device.

        Returns:
            Baudrate of the drive.
        """
        net = self.mc._get_network(servo)
        if isinstance(net, CanopenNetwork):
            return CanBaudrate(net.baudrate)
        raise IMError(f"The servo {servo} is not a CANopen device.")

    def get_ip(self, servo: str = DEFAULT_SERVO) -> str:
        """Get the IP for Ethernet communications.

        Args:
            servo: alias of the servo.

        Raises:
            IMError: If the servo is not an Ethernet device.

        Returns:
            IP of the drive.
        """
        drive = self.mc._get_drive(servo)
        net = self.mc._get_network(servo)
        if isinstance(net, EthernetNetwork):
            return str(drive.target)
        else:
            raise IMError("You need an Ethernet communication to use this function")

    def get_slave_id(self, servo: str = DEFAULT_SERVO) -> int:
        """Get the EtherCAT slave ID of a given servo.

        Args:
            servo: alias of the servo.

        Raises:
            IMError: If the servo is not an EtherCAT slave.

        Returns:
            Slave ID of the servo.

        """
        drive = self.mc._get_drive(servo)
        net = self.mc._get_network(servo)
        if isinstance(net, EoENetwork) and isinstance(drive.target, str):
            return net._configured_slaves[drive.target]
        elif isinstance(net, EthercatNetwork) and isinstance(drive.target, int):
            return drive.target
        raise IMError(f"The servo {servo} is not an EtherCAT slave.")

    def get_name(self, servo: str = DEFAULT_SERVO) -> str:
        """Get the drive's name.

        Args:
            servo: Alias of the servo.

        Returns:
            The name of the drive.
        """
        drive = self.mc._get_drive(servo)
        drive_name = drive.name
        return f"{drive_name}"

    def get_communication_type(self, servo: str = DEFAULT_SERVO) -> CommunicationType:
        """Get the connected drive's communication type.

        Args:
            servo: alias of the connected drive.

        Returns:
            CANopen, Ethernet, or EtherCAT.

        """
        drive_network = self.mc._get_network(servo)
        if isinstance(drive_network, CanopenNetwork):
            communication_type = CommunicationType.Canopen
        elif isinstance(drive_network, EthernetNetwork):
            communication_type = CommunicationType.Ethernet
        elif isinstance(drive_network, (EoENetwork, EthercatNetwork)):
            communication_type = CommunicationType.Ethercat
        else:
            raise NotImplementedError("Unknown communication type.")
        return communication_type

    def get_full_name(self, servo: str = DEFAULT_SERVO) -> str:
        """Return the full name of the drive [Product name] [Name] ([Target]).

        Args:
            servo: Drive alias.

        Returns:
            Full name.
        """
        prod_name = self.get_product_name(servo)
        name = self.get_name(servo)
        full_name = f"{prod_name} - {name}"
        net = self.mc._get_network(servo)
        if isinstance(net, EthernetNetwork):
            ip = self.get_ip(servo)
            full_name = f"{full_name} ({ip})"
        return full_name

    def get_subnodes(self, servo: str = DEFAULT_SERVO) -> dict[int, SubnodeType]:
        """Return a dictionary with the subnodes IDs as keys and their type as values.

        Args:
            servo: Drive alias.

        Returns:
            Dictionary of subnode ids and their type.
        """
        drive = self.mc._get_drive(servo)
        return drive.subnodes

    def get_categories(self, servo: str = DEFAULT_SERVO) -> dict[str, str]:
        """Return dictionary categories instance.

        Args:
            servo: Drive alias.

        Raises:
            IMError: If dictionary categories are not defined.

        Returns:
            Categories instance.
        """
        drive = self.mc._get_drive(servo)
        dictionary_categories = drive.dictionary.categories
        if not dictionary_categories:
            raise IMError("Dictionary categories are not defined.")
        category_ids = dictionary_categories.category_ids
        categories: dict[str, str] = {}
        for cat_id in category_ids:
            categories[cat_id] = dictionary_categories.labels(cat_id)["en_US"]
        return categories

    def get_dictionary_file_name(self, servo: str = DEFAULT_SERVO) -> str:
        """Return dictionary file name.

        Args:
            servo: Drive alias.

        Returns:
            Dictionary file name.
        """
        drive = self.mc._get_drive(servo)
        return str(os.path.basename(drive.dictionary.path))

    def get_encoded_image_from_dictionary(self, servo: str = DEFAULT_SERVO) -> Optional[str]:
        """Get the encoded product image from a drive dictionary.

        This function reads a dictionary of a drive, and it parses whether the dictionary file has a
        DriveImage tag and its content.

        Args:
            servo: Alias of the drive.

        Returns:
            The encoded image or NoneType object.
        """
        drive = self.mc._get_drive(servo)
        return drive.dictionary.image

    def is_comkit(self, servo: str = DEFAULT_SERVO) -> bool:
        """Check if drive is connected via a COM-KIT.

        Args:
            servo: Alias of the drive.

        Returns:
            True if using COM-KIT, False otherwise.

        """
        drive = self.mc._get_drive(servo)
        return drive.dictionary.coco_product_code == self.PRODUCT_CODE_COMKIT



import contextlib
import typing
from functools import wraps
from queue import Empty, Full, Queue
from typing import Callable


class StopExceptionError(Exception):
    """Stop exception."""


T = typing.TypeVar("T")


class Stoppable:
    """Stoppable class.

    It allows a test to be stoppable.

    """

    stop_queue: Queue[StopExceptionError] = Queue(1)

    @staticmethod
    def stoppable(fun: Callable[..., T]) -> Callable[..., T]:
        """Stoppable decorator.

        Args:
            fun: The function to decorate.

        Returns:
            The decorated function.

        """

        @wraps(fun)
        def wrapper(self, *args, **kwargs):  # type: ignore[no-untyped-def]
            self.check_stop()
            return fun(self, *args, **kwargs)

        return wrapper

    def reset_stop(self) -> None:
        """Reset the stop."""
        with contextlib.suppress(Empty):
            self.stop_queue.get(block=False)

    def stop(self) -> None:
        """Stop the test."""
        with contextlib.suppress(Full):
            self.stop_queue.put(StopExceptionError(), block=False)

    def check_stop(self) -> None:
        """Check if the test was stopped."""
        try:
            stop_exception = self.stop_queue.get(block=False)
        except Empty:
            pass
        else:
            raise stop_exception

    def stoppable_sleep(self, timeout: float) -> None:
        """A stoppable sleep.

        Args:
            timeout: Time to sleep.

        """
        try:
            stop_exception = self.stop_queue.get(timeout=timeout)
        except Empty:
            pass
        else:
            raise stop_exception



from typing import TYPE_CHECKING, Union

from ingenialink import RegDtype
from typing_extensions import override

from ingeniamotion.fsoe_master.fsoe import FSoEApplicationParameter, FSoEDataType

if TYPE_CHECKING:
    from ingenialink.ethercat.register import EthercatRegister
    from ingenialink.ethercat.servo import EthercatServo

__all__ = ["SafetyParameter", "SafetyParameterDirectValidation"]

PARAM_VALUE_TYPE = Union[int, float, str, bytes]


class SafetyParameter:
    """Safety Parameter.

    Represents a parameter that modifies how the safety application works.

    Base class is used for modules that use SRA CRC Check mechanism.
    """

    def __init__(self, register: "EthercatRegister", servo: "EthercatServo"):
        self.__register = register
        self.__servo = servo

        self.__value = servo.read(register)

    @property
    def register(self) -> "EthercatRegister":
        """Get the register associated with the safety parameter."""
        return self.__register

    def get(self) -> PARAM_VALUE_TYPE:
        """Get the value of the safety parameter.

        Returns:
            The value of the safety parameter.
        """
        return self.__value

    def set(self, value: PARAM_VALUE_TYPE) -> None:
        """Set the value of the safety parameter."""
        self.__servo.write(self.__register, value)
        self.__value = value

    def is_mismatched(self) -> tuple[bool, PARAM_VALUE_TYPE, PARAM_VALUE_TYPE]:
        """Check if the safety parameter value is mismatched.

        Returns:
            A tuple with a boolean indicating if the value is mismatched,
            the master parameter value, and the value on the slave,
        """
        slave_value = self.__servo.read(self.__register)
        return self.__value != slave_value, self.__value, slave_value

    def set_without_updating(self, value: PARAM_VALUE_TYPE) -> None:
        """Set the value of the safety parameter without updating the drive internal value."""
        self.__value = value

    def set_to_slave(self) -> None:
        """Set the value of the safety parameter to the slave."""
        self.__servo.write(self.__register, self.__value)


IL_TO_FSOE_DATATYPES = {
    RegDtype.BOOL: FSoEDataType.BOOL,
    RegDtype.FLOAT: FSoEDataType.FLOAT,
    RegDtype.U8: FSoEDataType.UINT8,
    RegDtype.U16: FSoEDataType.UINT16,
    RegDtype.U32: FSoEDataType.UINT32,
    RegDtype.S8: FSoEDataType.INT8,
    RegDtype.S16: FSoEDataType.INT16,
    RegDtype.S32: FSoEDataType.INT32,
}


class SafetyParameterDirectValidation(SafetyParameter):
    """Safety Parameter with direct validation via FSoE.

    Safety Parameter that is validated directly via FSoE in application
     state instead of SRA CRC Check
    """

    def __init__(self, register: "EthercatRegister", drive: "EthercatServo"):
        super().__init__(register, drive)

        self.fsoe_application_parameter = FSoEApplicationParameter(
            name=register.identifier,
            initial_value=self.get(),
            data_type=IL_TO_FSOE_DATATYPES[register.dtype],
        )

    @override
    def set(self, value: Union[int, float, str, bytes]) -> None:
        super().set(value)
        self.fsoe_application_parameter.set(value)

    @override
    def set_without_updating(self, value: Union[int, float, str, bytes]) -> None:
        super().set_without_updating(value)
        self.fsoe_application_parameter.set(value)



__all__ = [
    "FSOE_MASTER_INSTALLED",
    "FSoEApplicationParameter",
    "FSoEDataType",
    "FSoEDictionary",
    "FSoEDictionaryItem",
    "FSoEDictionaryItemInput",
    "FSoEDictionaryItemInputOutput",
    "FSoEDictionaryItemOutput",
    "FSoEDictionaryMap",
    "FSoEDictionaryMappedItem",
    "BaseMasterHandler",
    "StateData",
    "State",
    "calculate_sra_crc",
    "align_bits",
]

try:
    from fsoe_master.fsoe_master import (
        ApplicationParameter as FSoEApplicationParameter,
    )
    from fsoe_master.fsoe_master import (
        DataType as FSoEDataType,
    )
    from fsoe_master.fsoe_master import (
        Dictionary as FSoEDictionary,
    )
    from fsoe_master.fsoe_master import (
        DictionaryItem as FSoEDictionaryItem,
    )
    from fsoe_master.fsoe_master import (
        DictionaryItemInput as FSoEDictionaryItemInput,
    )
    from fsoe_master.fsoe_master import (
        DictionaryItemInputOutput as FSoEDictionaryItemInputOutput,
    )
    from fsoe_master.fsoe_master import (
        DictionaryItemOutput as FSoEDictionaryItemOutput,
    )
    from fsoe_master.fsoe_master import (
        DictionaryMap as FSoEDictionaryMap,
    )
    from fsoe_master.fsoe_master import (
        DictionaryMappedItem as FSoEDictionaryMappedItem,
    )
    from fsoe_master.fsoe_master import (
        MasterHandler as BaseMasterHandler,
    )
    from fsoe_master.fsoe_master import State, StateData, align_bits, calculate_sra_crc


except ImportError:
    FSOE_MASTER_INSTALLED = False
else:
    FSOE_MASTER_INSTALLED = True



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, cast

from ingenialink.dictionary import Dictionary, DictionaryError

from ingeniamotion._utils import weak_lru

if TYPE_CHECKING:
    from ingenialink import Servo


class Error:
    """Class to represent an error from the servo."""

    def __init__(self, error_id: int, dictionary_error: Optional[DictionaryError] = None):
        """Constructor.

        Args:
            error_id: Error ID.
            dictionary_error: DictionaryError instance from the dictionary, if available.
        """
        self.__error_id = error_id
        self.__dictionary_error = dictionary_error

    @property
    def error_id(self) -> int:
        """Get the error ID."""
        return self.__error_id

    @property
    def error_description(self) -> str:
        """Get the error description."""
        if self.__dictionary_error is not None and self.__dictionary_error.description is not None:
            return self.__dictionary_error.description
        return f"Unknown error {self.error_id} / 0x{self.error_id:X}"

    @classmethod
    def from_id(cls, error_id: int, dictionary: Optional[Dictionary] = None) -> Optional["Error"]:
        """Get an Error instance from an error ID.

        Args:
            error_id: Error ID.
            dictionary: Dictionary to get the error description from.

        Returns:
            Error: Error instance, or None if error_id is 0.
        """
        if error_id == 0:
            return None

        dictionary_error = None

        if dictionary:
            dictionary_error = dictionary.errors.get(error_id, None)

        return cls(error_id, dictionary_error)

    def __repr__(self) -> str:
        """Get a string representation of the Error instance.

        Returns:
            str: String representation of the Error instance.
        """
        return (
            f"<Error object at {hex(id(self))} "
            f"error_id={self.error_id} "
            f"error_description='{self.error_description}'>"
        )


@dataclass()
class ErrorQueueDescriptor:
    """Descriptor for an error queue in a servo."""

    last_error_reg_uid: str
    total_error_reg_uid: str
    error_request_index_reg_uid: str
    error_request_code_reg_uid: str
    max_index_request: int


MCUA_ERROR_QUEUE = ErrorQueueDescriptor(
    last_error_reg_uid="FSOE_LAST_ERROR_MCUA",
    total_error_reg_uid="FSOE_TOTAL_ERROR_MCUA",
    error_request_index_reg_uid="FSOE_ERROR_REQUEST_INDEX_MCUA",
    error_request_code_reg_uid="FSOE_ERROR_REQUEST_CODE_MCUA",
    max_index_request=31,
)

MCUB_ERROR_QUEUE = ErrorQueueDescriptor(
    last_error_reg_uid="FSOE_LAST_ERROR_MCUB",
    total_error_reg_uid="FSOE_TOTAL_ERROR_MCUB",
    error_request_index_reg_uid="FSOE_ERROR_REQUEST_INDEX_MCUB",
    error_request_code_reg_uid="FSOE_ERROR_REQUEST_CODE_MCUB",
    max_index_request=31,
)


class ServoErrorQueue:
    """Class to manage a error queue of a servo."""

    def __init__(
        self,
        descriptor: ErrorQueueDescriptor,
        servo: "Servo",
    ):
        self.descriptor = descriptor
        self.__servo = servo
        self.__dictionary = servo.dictionary

        # Total number of errors that were last read to obtain pending errors
        self.__last_read_total_errors_pending = 0

    def __read_int_reg(self, reg_uid: str) -> int:
        return cast("int", self.__servo.read(reg_uid))

    def get_last_error(self) -> Optional[Error]:
        """Get the last error from the servo's error queue.

        Returns:
            Optional[Error]: The last error, or None if there is no error.
        """
        return Error.from_id(
            self.__read_int_reg(self.descriptor.last_error_reg_uid), self.__dictionary
        )

    def get_number_total_errors(self) -> int:
        """Get the total number of errors from the servo's error queue.

        Returns:
            int: Total number of errors.
        """
        return self.__read_int_reg(self.descriptor.total_error_reg_uid)

    @property
    @weak_lru()
    def max_number_of_errors_in_buffer(self) -> int:
        """Get the maximum number of errors in the buffer from the servo's error queue.

        If more errors occur. oldest ones are discarded.
        """
        return self.descriptor.max_index_request + 1

    def get_error_by_index(self, index: int) -> Optional[Error]:
        """Get the error from the servo's error queue.

        Args:
            index: Index of the error from the servo's error queue.

        Returns:
            The error at the given index, or None if there is no error.
        """
        self.__servo.write(self.descriptor.error_request_index_reg_uid, index)
        return Error.from_id(
            self.__read_int_reg(self.descriptor.error_request_code_reg_uid), self.__dictionary
        )

    def __get_pending_error_indexes(
        self, current_total_errors: int
    ) -> tuple[tuple[int, ...], bool]:
        """Get the indexes of the pending errors from the servo's error queue.

        Indicates the indexes of the error queue that are pending to be read since the last
        time it the queue was read, according to the total number of errors reported by the servo.

        It takes into account overflow and wrap around of the error queue.

        Also indicates if any errors were lost and will not be read.
        Happens when the n pending errors to read is higher than the queue length

        Returns:
            tuple with:
                tuple of indexes of the errors pending to be read
                boolean indicating if errors were lost
        """
        n_pending_errors = current_total_errors - self.__last_read_total_errors_pending
        errors_lost = n_pending_errors > self.max_number_of_errors_in_buffer

        first_n_error_to_read = self.__last_read_total_errors_pending
        if errors_lost:
            # Previous errors have been lost and can't be read
            first_n_error_to_read = current_total_errors - self.max_number_of_errors_in_buffer

        return tuple(
            idx % self.max_number_of_errors_in_buffer
            for idx in range(first_n_error_to_read, current_total_errors)
        ), errors_lost

    def get_pending_errors(self) -> tuple[list[Error], bool]:
        """Get the pending errors from the servo's error queue.

        Indicates the errors that have occurred since the last time this method was called.

        Returns:
            A tuple containing: List of pending errors, and a boolean indicating
                if any errors were lost due to buffer overflow.
        """
        total_errors = self.get_number_total_errors()
        pending_indexes, errors_lost = self.__get_pending_error_indexes(total_errors)
        errors = [
            error for idx in pending_indexes if (error := self.get_error_by_index(idx)) is not None
        ]

        self.__last_read_total_errors_pending = total_errors

        return errors, errors_lost



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from ingenialogger import get_logger
from typing_extensions import override

from ingeniamotion.fsoe_master.frame import FSoEFrame
from ingeniamotion.fsoe_master.fsoe import (
    FSoEDictionaryMap,
    FSoEDictionaryMappedItem,
    align_bits,
)
from ingeniamotion.fsoe_master.safety_functions import STOFunction

logger = get_logger(__name__)


class FSoEFrameRules(Enum):
    """FSoE frame rules."""

    OBJECTS_ALIGNED = auto()
    OBJECTS_IN_FRAME = auto()
    STO_COMMAND_FIRST = auto()
    SAFE_DATA_BLOCKS_VALID = auto()
    OBJECTS_SPLIT_RESTRICTED = auto()


@dataclass(frozen=True)
class InvalidFSoEFrameRule:
    """Information about an invalid FSoE frame rule."""

    rule: FSoEFrameRules  # The rule that was violated
    exception: str  # Description of the error
    items: list[FSoEDictionaryMappedItem]  # Dictionary item that caused the rule to be invalid


@dataclass(frozen=True)
class FSoEFrameRuleValidatorOutput:
    """Output of FSoE frame rule validation.

    Contains the validation result and any exceptions raised during validation.
    """

    rules: list[FSoEFrameRules]
    """List of FSoE frame rules that were validated."""
    exceptions: dict[FSoEFrameRules, InvalidFSoEFrameRule]
    """Exceptions raised during validation.

    If no exception was raised for a rule, it won't be present in the dictionary.
    """

    def is_rule_valid(self, rule: FSoEFrameRules) -> bool:
        """Check if a specific rule is valid.

        Args:
            rule: The FSoE frame rule to check.

        Returns:
            True if the rule is valid, False otherwise.

        Raises:
            ValueError: If the rule is not in the validated rules list.
        """
        if rule not in self.rules:
            raise ValueError(f"Rule {rule} is not in the validated rules list.")
        return rule not in self.exceptions


class FSoEFrameConstructionError(Exception):
    """FSoE frame construction exceptions."""

    _ERROR_MESSAGE: str = "Map validation failed with errors"

    def __init__(self, validation_errors: FSoEFrameRuleValidatorOutput) -> None:
        """Initialize with validation errors.

        Args:
            validation_errors: The validation output containing the rules and exceptions.
        """
        error_details = []
        for idx, (rule, exception) in enumerate(validation_errors.exceptions.items(), 1):
            error_detail = f"  {idx}. Rule: {rule}\n"
            error_detail += f"     Error: {exception.exception}\n"
            error_details.append(error_detail)
        full_message = f"{self._ERROR_MESSAGE}:\n" + "\n".join(error_details)
        super().__init__(full_message)


class FSoEFrameRuleValidator(ABC):
    """Base class for FSoE frame rule validators.

    Provides a common interface for validating FSoE frame rules.
    """

    def __init__(self, rules: list[FSoEFrameRules]) -> None:
        self.rules: list[FSoEFrameRules] = rules

    @abstractmethod
    def _validate(
        self, dictionary_map: FSoEDictionaryMap, rules: list[FSoEFrameRules]
    ) -> FSoEFrameRuleValidatorOutput:
        raise NotImplementedError

    def validate(
        self, dictionary_map: FSoEDictionaryMap, rules: Optional[list[FSoEFrameRules]] = None
    ) -> FSoEFrameRuleValidatorOutput:
        """Validate the FSoE frame rules.

        Args:
            dictionary_map: The dictionary map to validate.
            rules: Optional list of specific rules to validate. If None, all rules are validated.

        Returns:
            The output of the validation containing the rules and exceptions.

        Raises:
            ValueError: If no valid rules are provided for evaluation.
        """
        eval_rules: list[FSoEFrameRules] = (
            self.rules if rules is None else [rule for rule in rules if rule in self.rules]
        )
        if not len(eval_rules):
            raise ValueError("No valid rules to evaluate. Please provide a valid list of rules.")
        return self._validate(dictionary_map, eval_rules)


class SafeDataBlocksValidator(FSoEFrameRuleValidator):
    """Validator for safe data blocks rule: 1-8 blocks, each 16-bit.

    If the frame contains only one block, it may be either 8 bits or 16 bits.

    Each safe data block is followed by a CRC_N, where N is the block index starting from 0.
    """

    __MAX_FRAME_OBJECTS: int = 45

    def __init__(self) -> None:
        super().__init__(
            rules=[
                FSoEFrameRules.SAFE_DATA_BLOCKS_VALID,
                FSoEFrameRules.OBJECTS_IN_FRAME,
                FSoEFrameRules.OBJECTS_SPLIT_RESTRICTED,
            ]
        )

    @staticmethod
    def __get_bits_in_data_block(
        items: list[tuple[Optional[int], Optional["FSoEDictionaryMappedItem"]]],
    ) -> int:
        slot_size_bits = 0
        for bits_in_slot, item in items:
            if bits_in_slot is not None:
                slot_size_bits += bits_in_slot
            else:
                slot_size_bits += item.bits if item else 0
        return slot_size_bits

    def _validate_safe_data_blocks_size(
        self,
        safe_data_blocks: list[
            tuple[int, list[tuple[Optional[int], Optional["FSoEDictionaryMappedItem"]]]]
        ],
    ) -> dict[FSoEFrameRules, InvalidFSoEFrameRule]:
        """Validate the size of safe data blocks in the dictionary map.

        Note:
            The only way in which the safe data blocks can have invalid size is if there is a bug in
            the FSoEFrame.generate_slot_structure method, which should always return safe data
            blocks with the correct size.

        Args:
            dictionary_map: The dictionary map to validate.
            safe_data_blocks: The list of safe data blocks to validate.

        Returns:
            Dictionary containing any validation errors found.
        """
        n_safe_data_blocks = len(safe_data_blocks)

        # PDO map will be valid if the map is empty (check `test_empty_map_8_bits` test)
        if n_safe_data_blocks == 0:
            return {}

        # Frames may contain 1 to 8 blocks of safe data (payload)
        if n_safe_data_blocks > 8:
            return {
                FSoEFrameRules.SAFE_DATA_BLOCKS_VALID: InvalidFSoEFrameRule(
                    rule=FSoEFrameRules.SAFE_DATA_BLOCKS_VALID,
                    exception=f"Expected 1-8 safe data blocks, found {n_safe_data_blocks}",
                    items=[
                        item
                        for _, slot_items in safe_data_blocks
                        for _, item in slot_items
                        if item is not None
                    ],
                )
            }
        # Validate each safe data block
        for data_slot_i, slot_items in safe_data_blocks:
            slot_size_bits = SafeDataBlocksValidator.__get_bits_in_data_block(slot_items)
            # If the frame contains only one block, it may be either 8 bits or 16 bits
            if n_safe_data_blocks == 1:
                # It can be smaller -> it will be padded to 16 bits when creating the PDO frame
                if slot_size_bits > 16:
                    return {
                        FSoEFrameRules.SAFE_DATA_BLOCKS_VALID: InvalidFSoEFrameRule(
                            rule=FSoEFrameRules.SAFE_DATA_BLOCKS_VALID,
                            exception="Single safe data block must be 16 bits or less "
                            f"with padding. Found {slot_size_bits}",
                            items=[item for _, item in slot_items if item is not None],
                        )
                    }
            # Each safe data block must be 16 bits
            elif data_slot_i != n_safe_data_blocks - 1:
                if slot_size_bits != 16:
                    return {
                        FSoEFrameRules.SAFE_DATA_BLOCKS_VALID: InvalidFSoEFrameRule(
                            rule=FSoEFrameRules.SAFE_DATA_BLOCKS_VALID,
                            exception=f"Safe data block {data_slot_i} must be 16 bits. "
                            f"Found {slot_size_bits}",
                            items=[item for _, item in slot_items if item is not None],
                        )
                    }
            # It can be smaller, because it will be padded to 16 bits when creating the PDO frame
            elif slot_size_bits > 16:
                return {
                    FSoEFrameRules.SAFE_DATA_BLOCKS_VALID: InvalidFSoEFrameRule(
                        rule=FSoEFrameRules.SAFE_DATA_BLOCKS_VALID,
                        exception=f"Last safe data block {data_slot_i} must be 16 bits or less "
                        f"with padding. Found {slot_size_bits}",
                        items=[item for _, item in slot_items if item is not None],
                    )
                }
        return {}

    def _validate_objects_in_frame(
        self,
        dictionary_map: FSoEDictionaryMap,
        safe_data_blocks: list[
            tuple[int, list[tuple[Optional[int], Optional["FSoEDictionaryMappedItem"]]]]
        ],
    ) -> dict[FSoEFrameRules, InvalidFSoEFrameRule]:
        """Validate the number of objects in the frame.

        1 object used for CMD
        1 object used for CONN_ID
        1 objects used per register mapped
        1 object used for each safe data block CRC

        Args:
            dictionary_map: The dictionary map to validate.
            safe_data_blocks: The list of safe data blocks to validate.

        Returns:
            Dictionary containing any validation errors found.
        """
        n_crcs = len(safe_data_blocks)  # One CRC per safe data block
        n_registers = len(dictionary_map)

        n_objects = 1 + n_registers + n_crcs + 1  # CMD, registers, CRCs, CONN_ID
        if n_objects > self.__MAX_FRAME_OBJECTS:
            return {
                FSoEFrameRules.OBJECTS_IN_FRAME: InvalidFSoEFrameRule(
                    rule=FSoEFrameRules.OBJECTS_IN_FRAME,
                    exception=(
                        "Total objects in frame exceeds limit: "
                        f"{n_objects} > {self.__MAX_FRAME_OBJECTS}"
                    ),
                    items=[item for item in dictionary_map if item is not None],
                )
            }
        return {}

    def _validate_size_of_split_objects(
        self,
        safe_data_blocks: list[
            tuple[int, list[tuple[Optional[int], Optional["FSoEDictionaryMappedItem"]]]]
        ],
    ) -> dict[FSoEFrameRules, InvalidFSoEFrameRule]:
        """Validate that only 32-bit objects may be split across multiple safe data blocks.

        Args:
            dictionary_map: The dictionary map to validate.
            safe_data_blocks: The list of safe data blocks to validate.

        Returns:
            Dictionary containing any validation errors found.
        """
        for data_slot_i, slot_items in safe_data_blocks:
            for bits_in_slot, item in slot_items:
                # bits_in_slot == None and bits_in_slot == item.bits:
                #      the item is fully contained in the current safe data block
                # item == None: the item is a virtual padding block,
                # which fits in the current safe data block
                # item.item == None: the item is a padding block, it can be split
                if (
                    bits_in_slot is None
                    or item is None
                    or bits_in_slot == item.bits
                    or item.item is None
                ):
                    continue
                # If the item != 32 bits, it cannot be split across multiple safe data blocks
                # 16 bit is checked by ObjectsAlignedValidator
                if item.bits < 16:
                    return {
                        FSoEFrameRules.OBJECTS_SPLIT_RESTRICTED: InvalidFSoEFrameRule(
                            rule=FSoEFrameRules.OBJECTS_SPLIT_RESTRICTED,
                            exception=(
                                f"Make sure that 8 bit objects belong to the same data block. "
                                f"Data slot {data_slot_i} contains split object {item.item.name}."
                            ),
                            items=[item],
                        )
                    }
        return {}

    @override
    def _validate(
        self, dictionary_map: FSoEDictionaryMap, rules: list[FSoEFrameRules]
    ) -> FSoEFrameRuleValidatorOutput:
        safe_data_blocks = list(
            FSoEFrame.generate_slot_structure(dictionary_map, FSoEFrame._FSoEFrame__SLOT_WIDTH)  # type: ignore[attr-defined]
        )

        exceptions: dict[FSoEFrameRules, InvalidFSoEFrameRule] = {}
        if FSoEFrameRules.SAFE_DATA_BLOCKS_VALID in rules:
            exceptions.update(self._validate_safe_data_blocks_size(safe_data_blocks))
        if FSoEFrameRules.OBJECTS_IN_FRAME in rules:
            exceptions.update(self._validate_objects_in_frame(dictionary_map, safe_data_blocks))
        if FSoEFrameRules.OBJECTS_SPLIT_RESTRICTED in rules:
            exceptions.update(self._validate_size_of_split_objects(safe_data_blocks))
        return FSoEFrameRuleValidatorOutput(rules=rules, exceptions=exceptions)


class ObjectsAlignedValidator(FSoEFrameRuleValidator):
    """Validator for object alignment in FSoE frames.

    Validates that 16-bit and larger objects are word-aligned within the payload.
    """

    def __init__(self) -> None:
        super().__init__(rules=[FSoEFrameRules.OBJECTS_ALIGNED])

    @override
    def _validate(
        self, dictionary_map: FSoEDictionaryMap, rules: list[FSoEFrameRules]
    ) -> FSoEFrameRuleValidatorOutput:
        exceptions: dict[FSoEFrameRules, InvalidFSoEFrameRule] = {}
        for item in dictionary_map:
            # Ignore paddings for alignment check
            if item.item is None:
                continue
            if item.bits >= 16 and item.position_bits % 16 != 0:
                next_alignment = align_bits(item.position_bits, 16)
                exceptions[FSoEFrameRules.OBJECTS_ALIGNED] = InvalidFSoEFrameRule(
                    rule=FSoEFrameRules.OBJECTS_ALIGNED,
                    exception=(
                        "Objects larger than 16-bit must be word-aligned. "
                        f"Object '{item.item.name}' found at position {item.position_bits}, "
                        f"next alignment is at {next_alignment}."
                    ),
                    items=[item],
                )
        return FSoEFrameRuleValidatorOutput(rules=rules, exceptions=exceptions)


class STOCommandFirstValidator(FSoEFrameRuleValidator):
    """Validator for the STO command position in FSoE frames.

    Validates that the STO command is always mapped to the first position.
    """

    def __init__(self) -> None:
        super().__init__(rules=[FSoEFrameRules.STO_COMMAND_FIRST])

    @override
    def _validate(
        self, dictionary_map: FSoEDictionaryMap, rules: list[FSoEFrameRules]
    ) -> FSoEFrameRuleValidatorOutput:
        exceptions: dict[FSoEFrameRules, InvalidFSoEFrameRule] = {}
        if not len(dictionary_map):
            exceptions[FSoEFrameRules.STO_COMMAND_FIRST] = InvalidFSoEFrameRule(
                rule=FSoEFrameRules.STO_COMMAND_FIRST,
                exception="Map is empty, STO command must be mapped to the first position",
                items=[],
            )
        else:
            first_item = dictionary_map._items[0]
            if first_item.item is None or first_item.item.name != STOFunction.COMMAND_UID:
                exceptions[FSoEFrameRules.STO_COMMAND_FIRST] = InvalidFSoEFrameRule(
                    rule=FSoEFrameRules.STO_COMMAND_FIRST,
                    exception="STO command must be mapped to the first position",
                    items=[first_item] if first_item is not None else [],
                )
        return FSoEFrameRuleValidatorOutput(rules=rules, exceptions=exceptions)


class FSoEDictionaryMapValidator:
    """Validator for FSoE Dictionary Maps.

    Validates that the dictionary map follows the rules for FSoE frame construction.
    """

    def __init__(self) -> None:
        """Initialize the FSoEDictionaryMapValidator."""
        self.__validators: list[FSoEFrameRuleValidator] = [
            SafeDataBlocksValidator(),
            ObjectsAlignedValidator(),
            STOCommandFirstValidator(),
        ]

    def validate_dictionary_map_fsoe_frame_rules(
        self,
        dictionary_map: FSoEDictionaryMap,
        rules: Optional[list[FSoEFrameRules]] = None,
        raise_exceptions: bool = False,
    ) -> FSoEFrameRuleValidatorOutput:
        """Validate that the FSoE dictionary map follows FSoE frame construction rules.

        Args:
            dictionary_map: The dictionary map to validate.
            rules: List of specific rules to validate. If None, all rules are validated.
            raise_exceptions: If True, raises an exception if any rule is invalid.
                If False, returns the validation output with exceptions. Defaults to False.

        Returns:
            The output of the validation containing the rules and exceptions.

        Raises:
            FSoEFrameConstructionError: if raise_exceptions is True and any rule is invalid.
        """
        rules_to_validate = list(set(rules)) if rules is not None else list(FSoEFrameRules)
        exceptions: dict[FSoEFrameRules, InvalidFSoEFrameRule] = {}
        for validator in self.__validators:
            if all(rule not in validator.rules for rule in rules_to_validate):
                continue
            output = validator.validate(dictionary_map, rules_to_validate)
            exceptions.update(output.exceptions)

        result = FSoEFrameRuleValidatorOutput(rules=rules_to_validate, exceptions=exceptions)
        if raise_exceptions and result.exceptions:
            raise FSoEFrameConstructionError(validation_errors=result)
        return result



from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional, Union

import ingenialogger
import numpy as np
from ingenialink.enums.register import RegDtype
from ingenialink.exceptions import ILValueError
from numpy import ndarray
from numpy.typing import NDArray

from ingeniamotion.enums import MonitoringVersion
from ingeniamotion.exceptions import IMDisturbanceError, IMStatusWordError
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

# Constants for typing
TYPE_MAPPED_REGISTERS_ALL = dict[str, Union[str, int, list[float]]]
TYPE_MAPPED_REGISTERS_NAME_AXIS = dict[str, Union[str, int, RegDtype]]
TYPE_MAPPED_REGISTERS_DATA = dict[str, list[Union[int, float]]]
TYPE_MAPPED_REGISTERS_DATA_NO_KEY = list[Union[int, float]]


def check_disturbance_disabled(
    func: Callable[..., Union[int, float, None]],
) -> Callable[..., Union[int, float, None]]:
    """Decorator that checks if disturbance is disabled before calling a function.

    Args:
        func: The function to called.

    Returns:
        The decorated function.

    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        disturbance_enabled = self.mc.capture.is_disturbance_enabled(
            servo=self.servo, version=self._version
        )
        if disturbance_enabled:
            raise IMDisturbanceError("Disturbance is enabled")
        return func(self, *args, **kwargs)

    return wrapper


class Disturbance:
    """Class to configure a disturbance in a servo.

    Args:
        mc : MotionController instance.
        servo : servo alias to reference it. ``default`` by default.
    """

    DISTURBANCE_FREQUENCY_DIVIDER_REGISTER = "DIST_FREQ_DIV"
    DISTURBANCE_MAXIMUM_SAMPLE_SIZE_REGISTER = "DIST_MAX_SIZE"
    MONITORING_DISTURBANCE_STATUS_REGISTER = "MON_DIST_STATUS"

    MONITORING_STATUS_ENABLED_BIT = 0x1
    REGISTER_MAP_OFFSET = 0x800

    __data_type_size = {
        RegDtype.U8: 1,
        RegDtype.S8: 1,
        RegDtype.U16: 2,
        RegDtype.S16: 2,
        RegDtype.U32: 4,
        RegDtype.S32: 4,
        RegDtype.U64: 8,
        RegDtype.S64: 8,
        RegDtype.FLOAT: 4,
    }

    def __init__(self, mc: "MotionController", servo: str = DEFAULT_SERVO) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.mapped_registers: list[TYPE_MAPPED_REGISTERS_NAME_AXIS] = []
        self.sampling_freq: Optional[float] = None
        self._version = mc.capture._check_version(servo)
        self.logger = ingenialogger.get_logger(__name__, drive=mc.servo_name(servo))
        self.max_sample_number = mc.capture.disturbance_max_sample_size(servo)
        if self._version < MonitoringVersion.MONITORING_V3:
            try:
                self.mc.capture.mcb_synchronization(servo=servo)
            except IMStatusWordError:
                self.logger.warning(
                    "MCB could not be synchronized. Motor is enabled.", drive=mc.servo_name(servo)
                )

    @check_disturbance_disabled
    def set_frequency_divider(self, divider: int) -> float:
        """Function to define disturbance frequency with a prescaler.

        Frequency will be
        ``Position & velocity loop rate frequency / prescaler``,  see
        :func:`ingeniamotion.configuration.Configuration.get_position_and_velocity_loop_rate`
        to know about this frequency. Monitoring/Disturbance must be disabled.

        Args:
            divider : determines disturbance frequency. It must be ``1`` or higher.

        Returns:
            Sample period in seconds.

        Raises:
            ValueError: If divider is less than ``1``.
        """
        if divider < 1:
            raise ValueError("divider must be 1 or higher")
        position_velocity_loop_rate = self.mc.configuration.get_position_and_velocity_loop_rate(
            servo=self.servo
        )
        self.sampling_freq = round(position_velocity_loop_rate / divider, 2)
        self.mc.communication.set_register(
            self.DISTURBANCE_FREQUENCY_DIVIDER_REGISTER, divider, servo=self.servo, axis=0
        )
        return 1 / self.sampling_freq

    @check_disturbance_disabled
    def map_registers(
        self,
        registers: Union[TYPE_MAPPED_REGISTERS_NAME_AXIS, list[TYPE_MAPPED_REGISTERS_NAME_AXIS]],
    ) -> float:
        """Map registers to Disturbance. Disturbance must be disabled.

        Args:
            registers : registers to map.
                Each register must be a dict with two keys.

                .. code-block:: python

                    {
                        "name": "CL_POS_SET_POINT_VALUE",  # Register name.
                        "axis": 1  # Register axis.
                        # If it has no axis field, by default axis 1.
                    }

        Returns:
            Max number of samples

        Raises:
            IMDisturbanceError: If the registers is an empty list.
            IMDisturbanceError: If the register is not allowed to be mapped as
                a disturbance register.
            TypeError: If some parameter has a wrong type.
        """
        if len(registers) == 0:
            raise IMDisturbanceError("No registers to be mapped.")
        if not isinstance(registers, list):
            registers = [registers]
        drive = self.mc.servos[self.servo]
        drive.disturbance_remove_all_mapped_registers()
        total_sample_size = 0
        for ch_idx, channel in enumerate(registers):
            subnode = channel.get("axis", DEFAULT_AXIS)
            if not isinstance(subnode, int):
                raise TypeError("Subnode value has to be an integer")
            register = channel["name"]
            if not isinstance(register, str):
                raise TypeError("Register key has to be a string")
            register_obj = self.mc.info.register_info(register, subnode, servo=self.servo)
            dtype = register_obj.dtype
            if register_obj.monitoring is None:
                drive.disturbance_remove_all_mapped_registers()
                raise IMDisturbanceError(f"{register} can not be mapped as a disturbance register")
            channel["dtype"] = dtype
            drive.disturbance_set_mapped_register(
                channel=ch_idx, uid=register, size=self.__data_type_size[dtype], axis=subnode
            )
            self.mapped_registers.append(channel)
            total_sample_size += self.__data_type_size[dtype]
        return self.max_sample_number / total_sample_size

    @staticmethod
    def __registers_data_adapter(
        registers_data: Union[
            list[
                Union[
                    int,
                    float,
                    NDArray[np.int32],
                    NDArray[np.float32],
                    TYPE_MAPPED_REGISTERS_DATA_NO_KEY,
                ]
            ],
            NDArray[np.int32],
            NDArray[np.float32],
        ],
    ) -> list[TYPE_MAPPED_REGISTERS_DATA_NO_KEY]:
        if isinstance(registers_data, ndarray):
            registers_data = registers_data.tolist()
            return [registers_data]  # type: ignore [list-item]
        elif isinstance(registers_data, list) and all(
            isinstance(data, list) for data in registers_data
        ):
            return registers_data  # type: ignore [return-value]
        elif isinstance(registers_data, list) and all(
            isinstance(data, (int, float)) for data in registers_data
        ):
            return [registers_data]  # type: ignore [list-item]
        elif isinstance(registers_data, list) and all(
            isinstance(data, ndarray) for data in registers_data
        ):
            for i, x in enumerate(registers_data):
                if isinstance(x, ndarray):
                    registers_data[i] = x.tolist()
            return registers_data  # type: ignore [return-value]
        else:
            raise TypeError(
                "Registers data adapter doesn't have the correct type for its input argument"
            )

    @check_disturbance_disabled
    def write_disturbance_data(
        self,
        registers_data: Union[
            list[
                Union[
                    int,
                    float,
                    NDArray[np.int32],
                    NDArray[np.float32],
                    TYPE_MAPPED_REGISTERS_DATA_NO_KEY,
                ]
            ],
            NDArray[np.int32],
            NDArray[np.float32],
        ],
    ) -> None:
        """Write data in mapped registers. Disturbance must be disabled.

        Args:
            registers_data :
                data to write in disturbance. Registers should have same order
                as in :func:`map_registers`.

        Raises:
            IMDisturbanceError: If there are no mapped registers or the sampling frequency is not
                set yet.
            IMDisturbanceError: If buffer size is not enough for all the
                registers and samples.
        """
        if len(self.mapped_registers) == 0 or self.sampling_freq is None:
            raise IMDisturbanceError("Disturbance is not correctly configured yet")
        adapted_registers_data = self.__registers_data_adapter(registers_data)
        drive = self.mc.servos[self.servo]
        self.__check_buffer_size_is_enough(adapted_registers_data)
        idx_list = list(range(len(adapted_registers_data)))
        dtype_list = [RegDtype(x["dtype"]) for x in self.mapped_registers]
        if self._version >= MonitoringVersion.MONITORING_V3:
            drive.disturbance_remove_data()
        try:
            drive.disturbance_write_data(idx_list, dtype_list, adapted_registers_data)
        except ILValueError as e:
            raise IMDisturbanceError(e)

    def map_registers_and_write_data(
        self, registers: Union[TYPE_MAPPED_REGISTERS_ALL, list[TYPE_MAPPED_REGISTERS_ALL]]
    ) -> None:
        """Map registers to Disturbance and write data. Disturbance must be disabled.

        Args:
            registers : registers to map and write data.
                Each register must be a dict with three keys:

                .. code-block:: python

                    {
                        "name": "CL_POS_SET_POINT_VALUE",  # Register name.
                        "axis": 1,  # Register axis.
                        # If it has no axis field, by default axis 1.
                        "data": [0.0, 0.1, 0.2, ...]  # Data for load in this register
                    }

        Raises:
            IMDisturbanceError: If the register is not allowed to be mapped as a
                disturbance register.
            IMDisturbanceError: If buffer size is not enough for all the
                registers and samples.
        """
        if not isinstance(registers, list):
            registers = [registers]
        registers_keys = []
        registers_data = []
        for channel in registers:
            subnode = channel.get("axis", DEFAULT_AXIS)
            register = channel["name"]
            registers_keys.append({"axis": subnode, "name": register})
            registers_data.append(channel["data"])
        self.map_registers(registers_keys)
        self.write_disturbance_data(registers_data)

    def __check_buffer_size_is_enough(
        self, registers: list[TYPE_MAPPED_REGISTERS_DATA_NO_KEY]
    ) -> None:
        total_buffer_size = 0
        for ch_idx, data in enumerate(registers):
            dtype = self.mapped_registers[ch_idx]["dtype"]
            if not isinstance(dtype, RegDtype):
                continue
            total_buffer_size += self.__data_type_size[dtype] * len(data)
        if total_buffer_size > self.max_sample_number:
            raise IMDisturbanceError(
                "Number of samples is too high. "
                f"Demanded size: {total_buffer_size} bytes, "
                f"buffer max size: {self.max_sample_number} bytes."
            )
        self.logger.debug(
            "Demanded size: %d bytes, buffer max size: %d bytes.",
            total_buffer_size,
            self.max_sample_number,
        )



import warnings
from enum import Enum, EnumMeta, IntEnum
from typing import Any, TypeVar

from ingenialink import (
    CanBaudrate,
    CanDevice,
    RegAccess,
    RegDtype,
)

T = TypeVar("T", bound=type[Enum])

__all__ = [
    "CanBaudrate",
    "CanDevice",
    "RegAccess",
    "RegDtype",
]


def export(obj: T) -> T:
    """Decorator use to explicitly export a class.

    Returns:
        T: The class itself, added to the module's __all__ list.
    """
    __all__.append(obj.__name__)
    return obj


class MetaEnum(EnumMeta):
    def __contains__(cls, item: object) -> bool:
        """Checks if item exists.

        Returns:
            bool: True if item is a valid enum member, False otherwise.
        """
        try:
            cls(item)
        except ValueError:
            return False
        return True


@export
class OperationMode(IntEnum, metaclass=MetaEnum):
    """Operation Mode Enum."""

    VOLTAGE = 0x00
    CURRENT_AMPLIFIER = 0x01
    CURRENT = 0x02
    CYCLIC_CURRENT = 0x22
    VELOCITY = 0x03
    PROFILE_VELOCITY = 0x13
    CYCLIC_VELOCITY = 0x23
    POSITION = 0x04
    PROFILE_POSITION = 0x14
    CYCLIC_POSITION = 0x24
    PROFILE_POSITION_S_CURVE = 0x44
    INTERPOLATED_POSITION = 0xA4
    PVT = 0xB4
    HOMING = 0x113
    TORQUE = 0x05
    CYCLIC_TORQUE = 0x25


@export
class Protocol(IntEnum, metaclass=MetaEnum):
    """Communication protocol."""

    TCP = 1
    UDP = 2


@export
class HomingMode(IntEnum, metaclass=MetaEnum):
    """Homing modes."""

    CURRENT_POSITION = 0
    POSITIVE_LIMIT_SWITCH = 1
    NEGATIVE_LIMIT_SWITCH = 2
    POSITIVE_IDX_PULSE = 3
    NEGATIVE_IDX_PULSE = 4
    POSITIVE_LIMIT_SWITCH_IDX_PULSE = 5
    NEGATIVE_LIMIT_SWITCH_IDX_PULSE = 6


@export
class MonitoringSoCType(IntEnum, metaclass=MetaEnum):
    """Monitoring start of condition type."""

    TRIGGER_EVENT_AUTO = 0
    """No trigger"""
    TRIGGER_EVENT_FORCED = 1
    """Forced trigger"""
    TRIGGER_EVENT_EDGE = 2
    """Edge trigger"""


@export
class MonitoringSoCConfig(IntEnum, metaclass=MetaEnum):
    TRIGGER_CONFIG_RISING_OR_FALLING = 0
    """Rising or falling edge trigger"""
    TRIGGER_CONFIG_RISING = 1
    """Rising edge trigger"""
    TRIGGER_CONFIG_FALLING = 2
    """Falling edge trigger"""


@export
class MonitoringProcessStage(IntEnum, metaclass=MetaEnum):
    """Monitoring process stage."""

    INIT_STAGE = 0x0
    """Init stage"""
    FILLING_DELAY_DATA = 0x2
    """Filling delay data"""
    WAITING_FOR_TRIGGER = 0x4
    """Waiting for trigger"""
    DATA_ACQUISITION = 0x6
    """Data acquisition"""
    END_STAGE = 0x8
    """End stage"""


@export
class SensorType(IntEnum, metaclass=MetaEnum):
    """Summit series feedback type enum."""

    ABS1 = 1
    """Absolute encoder 1"""
    INTGEN = 3
    """Internal generator"""
    QEI = 4
    """Digital/Incremental encoder 1"""
    HALLS = 5
    """Digital halls"""
    SSI2 = 6
    """Secondary SSI"""
    BISSC2 = 7
    """Absolute encoder 2"""
    QEI2 = 8
    """Digital/Incremental encoder 2"""
    SINCOS = 10
    """SinCos encoder 1"""


@export
class SensorCategory(IntEnum, metaclass=MetaEnum):
    """Feedback category enum."""

    ABSOLUTE = 0
    INCREMENTAL = 1


@export
class PhasingMode(IntEnum, metaclass=MetaEnum):
    """Phasing modes."""

    NON_FORCED = 0
    """Non forced"""
    FORCED = 1
    """Forced"""
    NO_PHASING = 2
    """No phasing"""


@export
class GeneratorMode(IntEnum, metaclass=MetaEnum):
    """Generator modes."""

    CONSTANT = 0
    """Constant"""
    SAW_TOOTH = 1
    """Saw tooth"""
    SQUARE = 2
    """Square"""


@export
class MonitoringVersion(IntEnum, metaclass=MetaEnum):
    """Monitoring version."""

    MONITORING_V1 = 0
    """Monitoring V1 used for Everest 1.8.1 and older."""
    MONITORING_V2 = 1
    """Monitoring V2 used for Capitan and some custom low-power drivers."""
    MONITORING_V3 = 2
    """Monitoring V3 used for Everest and Capitan newer than 1.8.1."""


@export
class SeverityLevel(IntEnum, metaclass=MetaEnum):
    """Test result enum."""

    SUCCESS = 0
    WARNING = 1
    FAIL = 2


@export
class CommunicationType(IntEnum, metaclass=MetaEnum):
    Canopen = 0
    Ethernet = 1
    Ethercat = 2


@export
class FeedbackPolarity(IntEnum, metaclass=MetaEnum):
    """Feedback polarity enum."""

    NORMAL = 0
    REVERSED = 1


@export
class CommutationMode(IntEnum, metaclass=MetaEnum):
    """Commutation Mode Enum."""

    SINUSOIDAL = 0
    TRAPEZOIDAL = 1
    SINGLE_PHASE = 2


@export
class FilterType(IntEnum, metaclass=MetaEnum):
    """Biquad filter type."""

    DISABLED = 0
    """ Filter disabled """
    LOWPASS = 1
    """ Low-pass filter """
    HIGHPASS = 2
    """ High-pass filter """
    BANDPASS = 3
    """ Band-pass filter """
    PEAK = 4
    """ Peak filter """
    NOTCH = 5
    """ Notch filter """
    LOWSHELF = 6
    """ Low Shelf filter """
    HIGHSHELF = 7
    """ High Shelf filter """


@export
class FilterSignal(Enum):
    """Signal to configure filter."""

    POSITION_FEEDBACK = "POS_FBK"
    """Position feedback."""
    POSITION_REFERENCE = "POS_REF"
    """Position reference."""
    VELOCITY_FEEDBACK = "VEL_FBK"
    """Velocity feedback."""
    VELOCITY_REFERENCE = "VEL_REF"
    """Velocity reference."""
    CURRENT_FEEDBACK = "CUR_FBK"
    """Current feedback."""
    CURRENT_REFERENCE = "CUR_REF"
    """Current reference."""


@export
class FilterNumber(IntEnum, metaclass=MetaEnum):
    """Filter number (1 or 2)."""

    FILTER1 = 1
    FILTER2 = 2


@export
class DigitalVoltageLevel(IntEnum, metaclass=MetaEnum):
    """GPIOs voltage level (HIGH/LOW) enum."""

    HIGH = 1
    LOW = 0


@export
class GPIOPolarity(IntEnum, metaclass=MetaEnum):
    """GPIOs polarity enum."""

    NORMAL = 0
    REVERSED = 1


@export
class GPI(IntEnum, metaclass=MetaEnum):
    """GPIs identifier enum."""

    GPI1 = 1
    GPI2 = 2
    GPI3 = 3
    GPI4 = 4


@export
class GPO(IntEnum, metaclass=MetaEnum):
    """GPOs identifier enum."""

    GPO1 = 1
    GPO2 = 2
    GPO3 = 3
    GPO4 = 4


@export
class FSoEState(IntEnum, metaclass=MetaEnum):
    """FSoE Master Handler state."""

    RESET = 0
    SESSION = 1
    CONNECTION = 2
    PARAMETER = 3
    DATA = 4


class TemperatureSensor(IntEnum, metaclass=MetaEnum):
    """Temperature sensor types enum."""

    NTC = 0
    """Negative Temperature Coefficient thermistor"""
    PTC = 1
    """Postive Temperature Coefficient thermistor"""
    LINEAR = 2
    """Linear thermistor"""
    RTD = 3
    """Resistance Temperature Detector"""
    SILICON = 4
    """Silicon Temperature sensor"""
    SWITCH = 5
    """Temperature Switch"""
    NONE = 6
    """No Temperature sensor"""


@export
class STOAbnormalLatchedStatus(IntEnum, metaclass=MetaEnum):
    """STO Abnormal Latched Status enum."""

    NOT_LATCHED = 0
    LATCHED = 1
    UNDETERMINATED = 2


# WARNING: Deprecated aliases
_DEPRECATED = {
    "COMMUNICATION_TYPE": "CommunicationType",
    "CAN_DEVICE": "CanDevice",
    "CAN_BAUDRATE": "CanBaudrate",
    "REG_DTYPE": "RegDtype",
    "REG_ACCESS": "RegAccess",
}


def __getattr__(name: str) -> Any:
    if name in _DEPRECATED:
        warnings.warn(
            f"{name} is deprecated, use {_DEPRECATED[name]} instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[_DEPRECATED[name]]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from ingenialink.canopen.register import CanopenRegister
from ingenialink.ethercat.dictionary import EthercatDictionaryV2
from ingenialink.pdo import PDOMap, PDOMapItem

from ingeniamotion.fsoe_master.safety_functions import (
    SafeInputsFunction,
    SS1Function,
    STOFunction,
)

if TYPE_CHECKING:
    from ingenialink.dictionary import Dictionary

    from ingeniamotion.fsoe_master.fsoe import (
        FSoEDictionaryItem,
        FSoEDictionaryMap,
        FSoEDictionaryMappedItem,
    )


@dataclass()
class FSoEFrameElements:
    """FSoE Frame Elements.

    Indicates uids of elements that compose the FSoE frame, excluding the safe data.
    """

    command_uid: str
    crcs_prefix: str
    connection_id_uid: str

    def get_crc_uid(self, data_slot_i: int) -> str:
        """Get the CRC element name for the given data slot index.

        Returns:
            The CRC element name for the given data slot index.
        """
        return f"{self.crcs_prefix}{data_slot_i}"


MASTER_FRAME_ELEMENTS = FSoEFrameElements(
    command_uid="FSOE_MASTER_FRAME_ELEM_CMD",
    crcs_prefix="FSOE_MASTER_FRAME_ELEM_CRC",
    connection_id_uid="FSOE_MASTER_FRAME_ELEM_CONNID",
)


SLAVE_FRAME_ELEMENTS = FSoEFrameElements(
    command_uid="FSOE_SLAVE_FRAME_ELEM_CMD",
    crcs_prefix="FSOE_SLAVE_FRAME_ELEM_CRC",
    connection_id_uid="FSOE_SLAVE_FRAME_ELEM_CONNID",
)


class FSoEFrame:
    """Class to build FSoE frames."""

    __SLOT_WIDTH = 16
    """Number of bits in a data slot of the Safety PDU."""

    @staticmethod
    def __create_pdo_item(
        servo_dictionary: "Dictionary",
        uid: str,
        item_type: type[PDOMapItem],
        size_bits: Optional[int] = None,
    ) -> PDOMapItem:
        if isinstance(servo_dictionary, EthercatDictionaryV2):
            # Dictionary V2 only supports SaCo phase 1
            # That does do not configure the pdo maps.
            # Padding items are enough to fill the map
            if uid == MASTER_FRAME_ELEMENTS.command_uid or uid == SLAVE_FRAME_ELEMENTS.command_uid:
                return item_type(size_bits=8)
            if (
                uid == MASTER_FRAME_ELEMENTS.connection_id_uid
                or uid == SLAVE_FRAME_ELEMENTS.connection_id_uid
            ):
                return item_type(size_bits=16)
            if uid.startswith((
                MASTER_FRAME_ELEMENTS.crcs_prefix,
                SLAVE_FRAME_ELEMENTS.crcs_prefix,
            )):
                return item_type(size_bits=16)
            if uid in [
                STOFunction.COMMAND_UID,
                SS1Function.COMMAND_UID.format(i=1),
                SafeInputsFunction.SAFE_INPUTS_UID,
            ]:
                return item_type(size_bits=1)

            raise NotImplementedError(f"Unknown FSoE item UID for Dictionary V2: {uid}")
        else:
            reg = servo_dictionary.get_register(uid)
            if not isinstance(reg, CanopenRegister):
                # Type could be narrowed to EthercatRegister
                # After this bugfix:
                # https://novantamotion.atlassian.net/browse/INGK-1111
                raise TypeError
            return item_type(reg, size_bits)

    @staticmethod
    def __create_pdo_safe_data_item(
        servo_dictionary: "Dictionary",
        item: "FSoEDictionaryItem",
        item_type: type[PDOMapItem],
        size_bits: Optional[int] = None,
    ) -> PDOMapItem:
        if item.item is None:
            # Padding item
            return item_type(size_bits=size_bits if size_bits is not None else item.bits)
        else:
            # I/O item
            return FSoEFrame.__create_pdo_item(
                servo_dictionary, item.item.name, item_type, size_bits=size_bits
            )

    @staticmethod
    def __get_crc_item(
        data_slot_i: int,
        frame_elements: FSoEFrameElements,
        pdo_item_type: type[PDOMapItem],
        servo_dictionary: "Dictionary",
    ) -> PDOMapItem:
        try:
            return FSoEFrame.__create_pdo_item(
                servo_dictionary, frame_elements.get_crc_uid(data_slot_i), pdo_item_type
            )
        except KeyError as e:
            raise NotImplementedError(
                f"No CRC found for data slot {data_slot_i}. Probably the PDU Map is wide"
            ) from e

    @staticmethod
    def generate_slot_structure(
        dict_map: "FSoEDictionaryMap", slot_width: int
    ) -> Iterator[tuple[int, list[tuple[Optional[int], Optional["FSoEDictionaryMappedItem"]]]]]:
        """Generate the slot structure for a dictionary map.

        Args:
            dict_map: The dictionary map to generate the slot structure for.
            slot_width: The width of each data slot in bits.

        Yields:
            Tuples of (slot_index, (bits_in_slot, item))
            If item is None, it means a virtual padding item.
            If bits_in_slot is None, it means that the item fits in the slot without padding.
        """
        data_slot_i = 0

        # The minimum bits for the initial data slot is 8 bits
        slot_bit_maximum = 8

        # List of items that will be in the current slot
        # (bits in the slot, item)
        current_slot_items: list[tuple[Optional[int], Optional[FSoEDictionaryMappedItem]]] = []

        for item in dict_map:
            if slot_bit_maximum == 8 and item.position_bits + item.bits >= slot_bit_maximum:
                # Since there's enough data to overflow the initial slot of 8 bits,
                # it will be of 16 bits instead
                slot_bit_maximum = slot_width

            if item.position_bits >= slot_bit_maximum:
                # This item must go in the next data slot
                if current_slot_items:
                    yield (data_slot_i, current_slot_items)
                    current_slot_items = []
                data_slot_i += 1
                slot_bit_maximum += slot_width

            if item.position_bits + item.bits <= slot_bit_maximum:
                # The item fits in the current slot
                current_slot_items.append((None, item))
            else:
                # The item must go in the current slot, and on the next one
                # Have a virtual padding with the remaining bits
                # As described on ETG5120 Section 5.3.3

                # Number of bits that will be used in the current slot,
                # taking into account that it may start in the middle of the slot
                item_bits_in_slot = slot_width - item.position_bits % slot_width
                current_slot_items.append((item_bits_in_slot, item))

                # Yield current slot
                yield (data_slot_i, current_slot_items)
                current_slot_items = []

                # There are remaining bits that must be mapped into virtual paddings
                remaining_bits_to_map = item.bits - item_bits_in_slot

                while remaining_bits_to_map > 0:
                    data_slot_i += 1
                    slot_bit_maximum += slot_width
                    bits_to_map_in_this_slot = min(remaining_bits_to_map, slot_width)
                    # Virtual Padding item
                    current_slot_items.append((bits_to_map_in_this_slot, None))
                    if remaining_bits_to_map > slot_width:
                        yield (data_slot_i, current_slot_items)
                        current_slot_items = []

                    remaining_bits_to_map -= bits_to_map_in_this_slot

        # Yield the last slot if it has items
        if current_slot_items:
            yield (data_slot_i, current_slot_items)

    @staticmethod
    def fill_pdo_map(
        dict_map: "FSoEDictionaryMap",
        servo_dictionary: "Dictionary",
        pdo_map: PDOMap,
        pdo_item_type: type[PDOMapItem],
        frame_elements: FSoEFrameElements,
    ) -> None:
        """Fill the PDO map with the FSoE frame elements and the safe data.

        Args:
            dict_map: The dictionary map to fill the PDO map with.
            servo_dictionary: The servo dictionary to use for the PDO items.
            pdo_map: The PDO map to fill.
            pdo_item_type: The type of the PDO items to create.
            frame_elements: The frame elements to use for the PDO map.
        """
        # Remove any existing items in the PDOMap
        pdo_map.clear()

        # Initial FSoE command
        pdo_map.add_item(
            FSoEFrame.__create_pdo_item(servo_dictionary, frame_elements.command_uid, pdo_item_type)
        )

        for data_slot_i, slot_items in FSoEFrame.generate_slot_structure(
            dict_map, FSoEFrame.__SLOT_WIDTH
        ):
            for bits_in_slot, item in slot_items:
                # Virtual padding item
                if item is None:
                    pdo_map.add_item(pdo_item_type(size_bits=bits_in_slot))
                else:
                    pdo_map.add_item(
                        FSoEFrame.__create_pdo_safe_data_item(
                            servo_dictionary, item, pdo_item_type, size_bits=bits_in_slot
                        )
                    )

            # Add CRC for this slot
            pdo_map.add_item(
                FSoEFrame.__get_crc_item(
                    data_slot_i, frame_elements, pdo_item_type, servo_dictionary
                )
            )

        pdo_map.add_item(
            FSoEFrame.__create_pdo_item(
                servo_dictionary, frame_elements.connection_id_uid, pdo_item_type
            )
        )
